import warnings

import click
import json
from abc import ABC, abstractmethod
from yaml import load as yaml_load
from dsdl.exception import DefineSyntaxError, DSDLImportError
from dsdl.warning import DuplicateDefineWarning, DefineSyntaxWarning
import os
from dataclasses import dataclass, field
from enum import Enum
from typing import Union, Optional
from collections import defaultdict
from .utils import *
from .parse_params import ParserParam
from .parse_field import ParserField, EleStruct
from .parse_class import ParserClass, EleClass

try:
    from yaml import CSafeLoader as YAMLSafeLoader
except ImportError:
    from yaml import SafeLoader as YAMLSafeLoader

CHECK_LOG = CheckLog(def_name="all", flag=0)


class Parser(ABC):
    """
    定义一个Parser的抽象基类（没法实例化的）：
    用处： 强制子类实现某些方法：写框架时常用。
    """

    @abstractmethod
    def _parse(self, data_file: str, library_path: str):
        """
        将yaml文件中的模型（struct）和标签(label)部分校验之后读入某个变量中
        """
        pass

    @abstractmethod
    def _generate(self) -> Optional[str]:
        """
        将内存里面的模型（struct）和标签(label)部分输出成ORM模型（python代码）
        """
        pass

    def process(self, data_file, library_path, output_file):
        self._parse(data_file, library_path)
        dsdl_py = self._generate()
        return dsdl_py


class TypeEnum(Enum):
    CLASS_DOMAIN = "class_domain"
    STRUCT = "struct"


@dataclass()
class StructORClassDomain:
    name: str
    type: TypeEnum = TypeEnum.STRUCT
    field_list: List[Union[EleStruct, EleClass]] = field(default_factory=list)
    parent: List[str] = None  # class_dom 特有的super category
    skeleton: List[List[int]] = None

    def __post_init__(self):
        """
        按照规定struct和class dom的名字不能是白皮书中已经包含的类型名，如List这些内定的名字
        """
        if self.name in TYPES_ALL:
            raise ValidationError(
                f"{self.name} is dsdl build-in value name, please rename it."
                f"Build-in value names are: {','.join(TYPES_ALL)}"
            )
        if self.name in [i + "Field" for i in TYPES_ALL]:
            raise ValidationError(
                f"{self.name} is dsdl build-in value name, please rename it."
                f"Build-in value names are: {','.join(TYPES_ALL)}"
            )
        check_name_format(self.name)


class DSDLParser(Parser, ABC):
    def __init__(self, report_flag: bool):
        self.struct_name = set()
        self.define_map = defaultdict(StructORClassDomain)
        self.dsdl_version = None
        self.meta = dict()
        self.report_flag = report_flag

    def _parse(self, data_file: str, library_path: str):
        """
        将yaml文件中的模型（struct）和标签(label)部分校验之后读入变量self.define_map中
            input_file_list: 读入的yaml文件
        """
        with open(data_file, "r") as f:
            desc = yaml_load(f, Loader=YAMLSafeLoader)

        # 校验必须有meta和$dsdl-version字段，见白皮书2.1
        try:
            self.dsdl_version = desc["$dsdl-version"]  # 存版本号，后续应该会使用（目前木有用）
        except KeyError as e:
            err_msg = f"data yaml must contains {e} section"
            if self.report_flag:
                temp_check_item = CheckLogItem(
                    def_name="all", msg=f"DefineSyntaxError: {err_msg}"
                )
                CHECK_LOG.sub_struct.append(temp_check_item)
                return
            raise DefineSyntaxError(err_msg)
        try:
            self.meta = desc["meta"]  # 存版meta信息，后续应该会使用（目前木有用）
        except KeyError as e:
            err_msg = f"data yaml must contains {e} section"
            if self.report_flag:
                temp_check_item = CheckLogItem(
                    def_name="all", msg=f"DefineSyntaxError: {err_msg}"
                )
                CHECK_LOG.sub_struct.append(temp_check_item)
                return
            raise DefineSyntaxError(err_msg)

        # 校验必须有data字段和data中的sample-type字段
        try:
            data_sample_type = desc["data"]["sample-type"]
        except KeyError as e:
            err_msg = (
                f"data yaml must contains {e} in `data` section with `sample-type`"
            )
            if self.report_flag:
                temp_check_item = CheckLogItem(
                    def_name="all", msg=f"DefineSyntaxError: {err_msg}"
                )
                CHECK_LOG.sub_struct.append(temp_check_item)
                return
            raise DefineSyntaxError(err_msg)

        # 校验必须有data字段和data中的sample-type字段
        try:
            global_info_type = desc["data"]["global-info-type"]
        except KeyError as e:
            global_info_type = None
            warning_msg = f"{e}, `global-info-type` is not defined"
            if self.report_flag:
                temp_check_item = CheckLogItem(
                    def_name="all", msg=f"DefineSyntaxWarning: {warning_msg}"
                )
                CHECK_LOG.sub_struct.append(temp_check_item)
            warnings.warn(warning_msg, DefineSyntaxWarning)

        # import原则：优先import `-p`指定的路径，木有报错；如果`-p`木有指定，先import 本地路径，没有的话import dsdl/library路径
        import_list = []
        if "$import" in desc:
            _import = desc["$import"]
            if library_path:
                for p in _import:
                    temp_p = os.path.join(library_path, p.strip() + ".yaml")
                    if os.path.exists(temp_p):
                        import_list.append(temp_p)
                    else:
                        raise DSDLImportError(
                            f"{p} is not exists in `{library_path}`,"
                            f"please given right path using `-p`."
                        )
            else:
                library_path = os.path.dirname(data_file)
                for p in _import:
                    temp_p = os.path.join(library_path, p.strip() + ".yaml")
                    if os.path.exists(temp_p):
                        import_list.append(temp_p)
                    else:
                        temp_p = os.path.join(
                            "dsdl", "dsdl_library", p.strip() + ".yaml"
                        )
                        if os.path.exists(temp_p):
                            import_list.append(temp_p)
                        else:
                            raise DSDLImportError(
                                f"{p} is not exists in neither `{library_path}` nor `dsdl/dsdl_library`,"
                                f"please given right path using `-p`."
                            )

        if "defs" in desc:
            # 获取yaml中模型（struct）和标签(label)部分的内容，存储在变量class_defi中，
            # 因为有不同格式的yaml(数据和模型放同一个yaml中或者分开放)，所以用if...else分别做处理
            # 注意区分这个root_class_defi,为啥要把他先存好？如果import的yaml中，有重复的模型，需要用它覆盖，参见白皮书2.5.1
            root_class_defi = desc["defs"].items()
            self.dsdl_version = desc["$dsdl-version"]  # 存版本号，后续应该会使用（目前木有用）
        else:
            root_class_defi = dict()

        import_desc = dict()
        for input_file in import_list:
            with open(input_file, "r") as f:
                import_desc.update(yaml_load(f, Loader=YAMLSafeLoader))

        if "defs" in import_desc:
            # 获取yaml中模型（struct）和标签(label)部分的内容，存储在变量class_defi中，
            # 因为有不同格式的yaml(数据和模型放同一个yaml中或者分开放)，所以用if...else分别做处理
            class_defi = import_desc["defs"]
        else:
            class_defi = import_desc
        # root_class_defi是数据yaml里面定义的模型，如果和import里面的重复了，会覆盖掉前面import的。参见白皮书2.5.1
        class_defi.update(root_class_defi)

        # get self.data_sample_type and self.sample_param_map
        try:
            PARAMS = ParserParam(
                sample_type=data_sample_type,
                struct_defi=class_defi,
                global_info_type=global_info_type,
            )
        except Exception as e:
            if self.report_flag:
                temp_check_item = CheckLogItem(
                    def_name=TypeEnum.STRUCT.value, msg=f"{e}"
                )
                CHECK_LOG.sub_struct.append(temp_check_item)
                return
            raise Exception(e)
        for define_name, define_value in class_defi.items():
            if define_name.startswith("$"):
                continue
            try:
                define_type = define_value["$def"]
            except KeyError as e:
                err_msg = f"{define_name} section must contains {e} sub-section"
                if self.report_flag:
                    temp_check_item = CheckLogItem(
                        def_name="all", msg=f"DefineSyntaxError: {err_msg}"
                    )
                    CHECK_LOG.sub_struct.append(temp_check_item)
                    return
                raise DefineSyntaxError(err_msg)
            if define_type == "struct" or define_type == "class_domain":
                if define_name in self.struct_name:
                    err_msg = f"{define_name} has defined."
                    if self.report_flag:
                        temp_check_item = CheckLogItem(
                            def_name="all", msg=f"DuplicateDefineWarning: {err_msg}"
                        )
                        CHECK_LOG.sub_struct.append(temp_check_item)
                        return
                    raise DuplicateDefineWarning(err_msg)
                self.struct_name.add(define_name)

        # loop for `class_defi` section，deal with each `struct` and `class_domain`
        for define_name, define_value in class_defi.items():
            if define_name.startswith("$"):
                # skip section like: $dsdl-version
                continue

            # each yaml file must contain '$def' section
            define_type = define_value["$def"]

            if define_type == "struct":
                define_info = StructORClassDomain(name=define_name)
                define_info.type = TypeEnum.STRUCT
                FIELD_PARSER = ParserField(self.struct_name)
                # verify each ele of `struct` in `class_defi`, and save in define_info
                struct_params = define_value.get("$params", None)
                field_list = dict()
                for raw_field in define_value["$fields"].items():
                    field_name = raw_field[0].strip()
                    field_type = raw_field[1].strip()
                    # 判断field_name是否为python保留字和是符合命名规范
                    try:
                        check_name_format(field_name)
                    except ValidationError as e:
                        if self.report_flag:
                            temp_check_item = CheckLogItem(
                                def_name=TypeEnum.STRUCT.value,
                                msg=f"Error in `{define_name}`, {e}",
                            )
                            CHECK_LOG.sub_struct.append(temp_check_item)
                            return
                        raise ValidationError(f"Error in `{define_name}`, {e}")
                    # 将参数实例化
                    if struct_params:
                        for param, value in PARAMS.general_param_map[
                            define_name
                        ].params_dict.items():
                            field_type = field_type.replace("$" + param, value)
                    try:
                        field_list[field_name] = EleStruct(
                            name=field_name,
                            type=FIELD_PARSER.pre_parse_struct_field(
                                field_name, field_type
                            ),
                        )
                    except Exception as e:
                        if self.report_flag:
                            temp_check_item = CheckLogItem(
                                def_name=TypeEnum.STRUCT.value, msg=f"{e}"
                            )
                            CHECK_LOG.sub_struct.append(temp_check_item)
                            return
                        raise Exception(e)
                # deal with `$optional` section after `$fields` section，
                # because we must ensure filed in `$optional` is the `filed_name` in `$fields` section.
                if "$optional" in define_value or FIELD_PARSER.optional:
                    optional_set = (
                        set(define_value["$optional"]) | FIELD_PARSER.optional
                    )
                    for optional_name in optional_set:
                        optional_name = optional_name.strip()
                        if optional_name in field_list:
                            temp_type = field_list[optional_name].type
                            temp_type = add_key_value_2_struct_field(
                                temp_type, "optional", True
                            )
                            field_list[optional_name].type = temp_type
                        else:
                            err_msg = (
                                f"Error in $optional: {optional_name} is not in $field"
                            )
                            if self.report_flag:
                                temp_check_item = CheckLogItem(
                                    def_name=TypeEnum.STRUCT.value,
                                    msg=f"DefineSyntaxError: {err_msg}",
                                )
                                CHECK_LOG.sub_struct.append(temp_check_item)
                                return
                            raise DefineSyntaxError(err_msg)
                for attr_name in FIELD_PARSER.is_attr:
                    temp_type = field_list[attr_name].type
                    temp_type = add_key_value_2_struct_field(temp_type, "is_attr", True)
                    field_list[attr_name].type = temp_type

                # get processed struct filed and save it in define_info
                if not field_list:
                    err_msg = "Struct must have fields more than or equal to 1"
                    if self.report_flag:
                        temp_check_item = CheckLogItem(
                            def_name=TypeEnum.STRUCT.value,
                            msg=f"DefineSyntaxError: {err_msg}",
                        )
                        CHECK_LOG.sub_struct.append(temp_check_item)
                        return
                    raise DefineSyntaxError(err_msg)
                define_info.field_list = list(field_list.values())

            elif define_type == "class_domain":
                try:
                    if "skeleton" in define_value:
                        CLASS_PARSER = ParserClass(
                            define_name,
                            define_value["classes"],
                            define_value["skeleton"],
                        )
                    else:
                        CLASS_PARSER = ParserClass(define_name, define_value["classes"])
                except Exception as e:
                    if self.report_flag:
                        temp_check_item = CheckLogItem(
                            def_name=TypeEnum.CLASS_DOMAIN.value, msg=f"{e}"
                        )
                        CHECK_LOG.sub_struct.append(temp_check_item)
                        return
                    raise e
                define_info = StructORClassDomain(name=CLASS_PARSER.class_name)
                # verify each ele (in other words: each label) of `class_domain`, and save in define_info
                define_info.type = TypeEnum.CLASS_DOMAIN
                define_info.field_list = CLASS_PARSER.class_field
                # define_info.parent = CLASS_PARSER.super_class_list
                define_info.skeleton = CLASS_PARSER.skeleton
            else:
                err_msg = f"error type {define_type} in yaml, type must be class_domain or struct."
                if self.report_flag:
                    temp_check_item = CheckLogItem(
                        def_name=TypeEnum.CLASS_DOMAIN.value,
                        msg=f"DefineSyntaxError: {err_msg}",
                    )
                    CHECK_LOG.sub_struct.append(temp_check_item)
                    return
                raise DefineSyntaxError(err_msg)

            self.define_map[define_info.name] = define_info

        CHECK_LOG.flag = 1

    def _generate(self) -> Optional[str]:
        """
        将内存里面的模型（struct）和标签(label)部分输出成ORM模型（python代码）
        """
        # check define cycles. 如果有环形（就是循环定义）那是不行滴～
        define_graph = nx.DiGraph()
        define_graph.add_nodes_from(self.define_map.keys())
        for key, val in self.define_map.items():
            if val.type == TypeEnum.STRUCT:
                for field_list in val.field_list:
                    for k in self.define_map.keys():
                        if k in field_list.type:
                            define_graph.add_edge(k, key)
            # elif val.type == TypeEnum.CLASS_DOMAIN:
            #     for field_list in val.parent:
            #         for k in self.define_map.keys():
            #             if k in field_list:
            #                 define_graph.add_edge(k, key)
        if not nx.is_directed_acyclic_graph(define_graph):
            err_msg = "define cycle found."
            if self.report_flag:
                temp_check_item = CheckLogItem(def_name="all", msg=f"{err_msg}")
                CHECK_LOG.sub_struct.append(temp_check_item)
                return
            raise err_msg

        dsdl_py = "# Generated by the dsdl parser. DO NOT EDIT!\n"
        dsdl_py += "from dsdl.types import *\n"
        dsdl_py += "from dsdl.geometry import *\n\n\n"
        ordered_keys = list(nx.topological_sort(define_graph))
        for idx, key in enumerate(ordered_keys):
            val = self.define_map[key]
            if val.type == TypeEnum.STRUCT:
                dsdl_py += f"class {key}(Struct):\n"
                for field_list in val.field_list:
                    dsdl_py += f"""    {field_list.name} = {field_list.type}\n"""
            if val.type == TypeEnum.CLASS_DOMAIN:
                dsdl_py += f"class {key}(ClassDomain):\n"
                dsdl_py += "    Classes = [\n"
                for ele_class in val.field_list:
                    if ele_class.super_categories:
                        temp = ", ".join(ele_class.super_categories)
                        dsdl_py += f"""        Label("{ele_class.label_value}", supercategories=[{temp}]),\n"""
                    else:
                        dsdl_py += f"""        Label("{ele_class.label_value}"),\n"""
                dsdl_py += "    ]\n"
                if val.skeleton:
                    dsdl_py += f"""    Skeleton = {val.skeleton}\n"""
            if idx != len(ordered_keys) - 1:
                dsdl_py += "\n\n"

        if self.report_flag:
            temp_check_item = CheckLogItem(
                def_name=TypeEnum.CLASS_DOMAIN.value,
                msg=f"success parser class dom",
                flag=1,
            )
            CHECK_LOG.sub_struct.append(temp_check_item)
            temp_check_item = CheckLogItem(
                def_name=TypeEnum.STRUCT.value, msg=f"success parser all strcut", flag=1
            )
            CHECK_LOG.sub_struct.append(temp_check_item)
            temp_check_item = CheckLogItem(
                def_name=TypeEnum.STRUCT.value,
                msg=f"success parser all parameters",
                flag=1,
            )
            CHECK_LOG.sub_struct.append(temp_check_item)
            CHECK_LOG.flag = 1
        return dsdl_py

    def process(
        self, data_file: str, library_path: str, output_file: str
    ) -> Optional[str]:
        self._parse(data_file, library_path)
        if self.report_flag and CHECK_LOG.flag:
            dsdl_py = self._generate()
        elif not self.report_flag:
            dsdl_py = self._generate()
        else:
            dsdl_py = None
        # if dsdl_py:
        #     print(
        #         f"Convert Yaml File to Python Code Successfully!\n"
        #         f"Yaml file (source): {data_file}\n"
        #         f"Output file (output): {output_file}"
        #     )
        if output_file:
            with open(output_file, "w") as of:
                print(dsdl_py, file=of)
        return dsdl_py


def dsdl_parse(
    dsdl_yaml: str,
    dsdl_library_path: str = None,
    output_file: str = None,
    report_flag: bool = False,
) -> Optional[str]:
    """
    Main function of parser yaml files to .py dsdl struct definition code.

    Arguments:
        dsdl_yaml: file path of `data definition yaml file`;
        dsdl_library_path: file path of '`$import` path' in `dsdl_yaml` file.
        output_file: output file path. if None, return string, else, generate .py file in output file path.
        report_flag: if return report

    Returns:
        Optional[str]: if output_file=None, return string of dsdl definition .py file;
                       else generate a .py file in `output_file` path.
    """
    dsdl_parser = DSDLParser(report_flag)
    res = dsdl_parser.process(dsdl_yaml, dsdl_library_path, output_file)
    return res


def check_dsdl_parser(
    dsdl_yaml: str,
    dsdl_library_path: str = None,
    output_file: str = None,
    report_flag: bool = False,
):
    dsdl_py = dsdl_parse(
        dsdl_yaml=dsdl_yaml,
        dsdl_library_path=dsdl_library_path,
        output_file=output_file,
        report_flag=report_flag,
    )
    res = {"dsdl_py": dsdl_py}
    if report_flag:
        res["check_log"] = json.dumps(CHECK_LOG.to_struct())
    return res


@click.command()
@click.option(
    "-y",
    "--yaml",
    "dsdl_yaml",
    type=str,
    required=True,
)
@click.option(
    "-p",
    "--path",
    "dsdl_library_path",
    type=str,
)
def parse(dsdl_yaml: str, dsdl_library_path: str = None):
    """
    a separate cli tool function for user to parser yaml files to .py dsdl struct definition code.

    Arguments:
        dsdl_yaml: file path of `data definition yaml file`;
        dsdl_library_path: file path of '`$import` path' in `dsdl_yaml` file.

    Returns:
        None: generate a .py file in the same folder of `dsdl_yaml` file.
    """
    dsdl_name = os.path.splitext(os.path.basename(dsdl_yaml))[0]
    output_file = os.path.join(os.path.dirname(dsdl_yaml), f"{dsdl_name}.py")
    dsdl_parse(dsdl_yaml, dsdl_library_path, output_file)



from typing import Optional, Dict, Tuple, List
import os
from collections import defaultdict
from .commons import Util
from ...geometry import LabelList, BBox
from ...types import Struct
from copy import deepcopy


class ImageSample:
    def __init__(self, image, palette):
        self.image = image
        self.palette = palette
        self.ground_truths = defaultdict(dict)

    def append_ground_truth(self, gt_path, gt_item):
        """
        {
            ./object/1:{
                bbox: {bb1: bbox_obj, bb2: bbox_obj}
                label: [l1: label_obj, l2: label_obj]
                bool: []
            }

            ./object/2: {
                bbox: {}
            }
        }
        """
        gt_dir, gt_name = os.path.split(gt_path)
        if hasattr(gt_item, "field_key"):
            field_key = gt_item.field_key.lower()
        else:
            field_key = gt_item.__class__.__name__.lower()
        if field_key not in self.ground_truths[gt_dir]:
            self.ground_truths[gt_dir][field_key] = {gt_name: gt_item}
        else:
            self.ground_truths[gt_dir][field_key][gt_name] = gt_item

    def visualize(self):
        image = self.image.to_image()
        image = image.convert("RGBA")
        image_label_lst = []  # store labels having no matched polygon or bbox, etc.
        for gt_dir, gt_item in self.ground_truths.items():
            field_keys = Util.sort_field(gt_item.keys())
            for field_key in field_keys:
                if field_key == "label":
                    image = LabelList(gt_item[field_key].values()).visualize(image=image,
                                                                             image_label_list=image_label_lst,
                                                                             palette=self.palette, **gt_item)
                else:
                    for ann_item in gt_item[field_key].values():
                        if hasattr(ann_item, "visualize"):
                            image = ann_item.visualize(image=image, palette=self.palette, **gt_item)
        LabelList(image_label_lst).visualize(image=image, palette=self.palette, bbox={"temp": BBox(0, 0, 0, 0)})
        return image

    def format(self):
        result = {"media": self.image, "ann": list(self.ground_truths.values())}
        return result


class ImageVisualizePipeline:
    """
    step 1:

    该类用于处理Dataset中输出的样本，指定field_list后，会先调用方法 self.whole_pipeline，提取各个字段的信息，举例：
    Dataset中的某个sample形如：
        sample = {
        "$image": {"img2": img_obj, "img1":img_obj},
        "$list": {"objects": [{"$image": {"img1": img_obj}, "$bbox": {"box": box_obj}}],
                  "object2": [{"$bbox":{"box": box_obj}}, {"$bbox":{"box": box_obj}}], },
        "$struct": {
                "struct1": {
                    "$image": {"img3": img_obj, "img4": img_obj}
                }
            }
        }
    指定的field_list内容为：[bbox, image]
    则self.whole_pipeline方法会将sample处理为：
    {
        ”image": {
            "./img2": img_obj,
            "./img1": img_obj,
            "./objects/0/img1": img_obj,
            "./struct1/img3": img_obj,
            "./struct1/img4": img_obj,
        },
        "bbox": {
            "./objects/0/box": box_obj,
            "./object2/0/box": box_obj,
            "./object2/1/box": box_obj
        }
    }

    step2:
    然后会调用self.group_media_and_ann方法，将图像与与之”目录路径“相同的标注进行组合，组合后图像和对应的标注会存储在一个 ImageVisualizer
    对象当中。ImageVisualizer类中还定义了可视化的相关方法，从而可以对每个图像&标注组合进行可视化。方法的返回值是一个字典，字典的键为图像路径，
    值即为ImageVisualizer对象，形如：
    {
        "./img2": ImageVisualizer_obj,
        "./img1": ImageVisualizer_obj,
        "./objects/0/img1": ImageVisualizer_obj,
    }

    step3:
    调用format/visualize方法，从而可以对各个ImageVisualizer对象进行文本/图像可视化。
    """
    PALETTE = {}
    BU_DIST_WEIGHT = 1
    TD_DIST_WEIGHT = 0.01
    DIST_TRHESH = 2

    def __init__(self, field_list: List[str], sample: Struct, palette: Optional[Dict[str, Tuple]] = None):
        self.field_list = field_list
        self.palette = self.PALETTE
        if palette is not None:  # 如果用户自己定义了label调色盘，则更新默认的调色盘
            self.palette = palette
        if "image" not in field_list:  # 因为是imagevisualizer类，所以field_list中必须包含image字段
            raise ValueError("'image' field not found in the field list to be visualized.")
        self.data_dic = sample.extract_field_info(field_list)
        self.visualize_result = self.group_media_and_ann()

    @classmethod
    def _match(cls, ann_path, image_paths):
        dists = [cls._calculate_distance(ann_path, image_path) for image_path in image_paths]
        min_dist = min(dists)
        res = [image_paths[_] for _ in range(len(image_paths)) if dists[_] == min_dist and dists[_] < cls.DIST_TRHESH]
        return res

    @classmethod
    def _calculate_distance(cls, ann_path, image_path):
        img_dir = os.path.split(image_path)[0]
        bu_distance = 0
        while not ann_path.startswith(img_dir + "/"):
            bu_distance += 1
            img_dir = os.path.split(img_dir)[0]
        td_distance = len(ann_path.replace(img_dir, "").split("/")) - 1
        return cls._metric(bu_distance, td_distance)

    @classmethod
    def _metric(cls, bu_distance, td_distance):
        return bu_distance * cls.BU_DIST_WEIGHT + td_distance * cls.TD_DIST_WEIGHT

    def group_media_and_ann(self):
        data_dic = self.data_dic
        image_dic = data_dic.pop("image")
        image_paths = list(image_dic.keys())
        result_dic = {k_: ImageSample(image_dic[k_], self.palette) for k_ in image_paths}
        for ann_dic in data_dic.values():
            for ann_path, ann_obj in ann_dic.items():
                matched_img_paths = self._match(ann_path, image_paths)
                for matched_img_path in matched_img_paths:
                    result_dic[matched_img_path].append_ground_truth(ann_path, ann_obj)
        return result_dic

    def visualize(self):
        vis_result = {}
        for image_path, sample_item in self.visualize_result.items():
            vis_result[image_path] = sample_item.visualize()
        return vis_result

    def format(self):
        format_result = []
        for image_vis_item in self.visualize_result.values():
            format_result.append(image_vis_item.format())
        return format_result



import json
import os
import os.path as osp

from ...types import Struct, StructMetaclass
from ...objectio import BaseFileReader


def check_struct(sample_type: StructMetaclass, sample: dict, file_reader: BaseFileReader):
    report = {"error_flag": False, "warning_flag": False, "normal_flag": True}

    try:
        struct_sample = sample_type(file_reader=file_reader, **sample)
    except Exception as e:
        report["error_msg"] = repr(e)
        report["error_flag"] = True
        report["normal_flag"] = False
        report["sample_content"] = sample
        return None, report

    field_not_matching_msg = []
    field_mappings = struct_sample.__mappings__
    struct_mappings = struct_sample.__struct_mappings__
    optional_mappings = struct_sample.__optional__
    for k in field_mappings:
        if k not in sample and k not in optional_mappings:
            warn_msg = f"Field not matching warning: required field '{k}' not found in sample."
            field_not_matching_msg.append(warn_msg)
            report["warning_flag"] = True
            report["normal_flag"] = False
    for k in struct_mappings:
        if k not in sample:
            warn_msg = f"Field not matching warning: required struct field '{k}' not found in sample."
            field_not_matching_msg.append(warn_msg)
            report["warning_flag"] = True
            report["normal_flag"] = False
    for k in sample:
        if k not in field_mappings and k not in struct_mappings:
            warn_msg = f"Field not matching warning: field key '{k}' in sample is not found in {sample_type.__name__} struct's fields."
            field_not_matching_msg.append(warn_msg)
            report["warning_flag"] = True
            report["normal_flag"] = False
        elif k in struct_mappings:
            struct_type = struct_mappings[k].__class__
            sample_value = sample[k]
            _, nested_report = check_struct(struct_type, sample_value, file_reader)
            if not nested_report["normal_flag"]:
                report["normal_flag"] = False
                report["warning_flag"] = report["warning_flag"] and nested_report["warning_flag"]
                if "warning_msgs" in nested_report:
                    field_not_matching_msg.extend(nested_report["warning_msgs"])
        elif k in field_mappings:
            this_field = field_mappings[k]
            if this_field.extract_key() != "$list":
                continue
            if isinstance(this_field.ele_type, Struct):
                struct_type = this_field.ele_type.__class__
                sample_value = sample[k]
                for s in sample_value:
                    _, nested_report = check_struct(struct_type, s, file_reader)
                    if not nested_report["normal_flag"]:
                        report["normal_flag"] = False
                        report["warning_flag"] = report["warning_flag"] and nested_report["warning_flag"]
                        if "warning_msgs" in nested_report:
                            field_not_matching_msg.extend(nested_report["warning_msgs"])
    if not report["normal_flag"]:
        report["sample_content"] = sample
    if field_not_matching_msg:
        report["warning_msgs"] = field_not_matching_msg

    return struct_sample, report


class Report:
    def __init__(self, output_path):
        self.parser_info = None
        self.sample_info = []
        self.global_info = None
        self.image_info = []
        self.output_path = output_path

    def set_global_info(self, item):
        self.global_info = item

    def add_sample_info(self, item):
        self.sample_info.append(item)

    def set_sample_info(self, info):
        self.sample_info = info

    def add_image_info(self, item):
        self.image_info.append(item)

    def set_parser_info(self, item):
        self.parser_info = item

    def parse_sample_info(self):
        work_dir = osp.split(self.output_path)[0]
        log_path = osp.join(work_dir, "samples_log.md")
        total_num, normal_num, warn_num, error_num = len(self.sample_info), 0, 0, 0
        with open(log_path, "w") as f:
            f.write("具体异常日志信息如下：" + os.linesep)
            f.write("```json\n")
            for idx, info in enumerate(self.sample_info):
                if info["normal_flag"]:
                    normal_num += int(info["normal_flag"])
                    # f.write(f"status SUCCESS" + os.linesep)
                    continue
                elif info["warning_flag"]:
                    f.write(f"sample {idx}: ")
                    warn_num += int(info["warning_flag"])
                    f.write(f"status WARNING" + "\n")
                    f.write("\twarning messages:" + "\n")
                    for msg in info["warning_msgs"]:
                        f.write("\t\t" + msg + "\n")
                    f.write("\tsample content:" + "\n")
                    f.write("\t\t" + f"{str(info['sample_content'])}" + "\n")
                elif info["error_flag"]:
                    f.write(f"sample {idx}: ")
                    error_num += int(info["error_flag"])
                    f.write("status ERROR" + "\n")
                    f.write("\terror messages:" + "\n")
                    f.write("\t\t" + f"{info['error_msg']}" + "\n")
                    f.write("\tsample content:" + "\n")
                    f.write("\t\t" + f"{str(info['sample_content'])}" + "\n")
            f.write("```" + os.linesep)
        return {"total": total_num, "error": error_num, "warn": warn_num, "normal": normal_num}

    def parse_image_info(self):
        work_dir, _ = osp.split(self.output_path)
        img_dir = osp.join(work_dir, "imgs")
        if not osp.exists(img_dir):
            os.mkdir(img_dir)
        res = []
        msgs = []
        statis = {"success": 0, "fail": 0}
        for idx, img_info in enumerate(self.image_info):
            success_flag = img_info["status"] == "success"
            im = img_info["img"]
            if success_flag:
                img_path = osp.join(img_dir, f"{idx}.png")
                rel_img_path = osp.join("imgs", f"{idx}.png")
                res.append(rel_img_path)
                im.save(img_path)
                msgs.append(f"image {idx} view success!")
                statis["success"] = statis["success"] + 1
            else:
                msgs.append(f"image {idx} view fail! detailed exception info is shown below:"
                            f"{os.linesep}" + img_info['msg'])
                statis["fail"] = statis["fail"] + 1
        return res, msgs, statis

    def generate_parser_info(self, file_handler):
        file_handler.write("## 1.解析器验证结果" + os.linesep)
        success_flag = self.parser_info["flag"] == 1
        file_handler.write(
            f"解析结果：{'<font color=green>SUCCESS</font>' if success_flag else '<font color=red>FAIL</font>'}" + os.linesep)
        file_handler.write("具体信息如下：" + os.linesep)
        file_handler.write("```python" + os.linesep)
        file_handler.write(json.dumps(self.parser_info, indent=2, ensure_ascii=False))
        file_handler.write(os.linesep + "```" + os.linesep)
        return success_flag

    def generate_sample_info(self, file_handler):
        if not self.sample_info:
            return False
        file_handler.write("## 2.样本验证结果" + os.linesep)
        if self.global_info:
            # write global_info check info
            file_handler.write("global-info验证结果如下：\n")
            if self.global_info["normal_flag"]:
                file_handler.write("global-info 实例化<font color=green>正常</font>\n")
            elif self.global_info["error_flag"]:
                file_handler.write("global-info 实例化<font color=red>失败</font>，报错信息如下：\n")
                file_handler.write(f"```python\n")
                file_handler.write(f"{self.global_info['error_msg']}\n")
                file_handler.write("```\n")
            elif self.global_info["warning_flag"]:
                file_handler.write("global-info 实例化<font color=orange>异常</font>，警报信息如下：\n")
                file_handler.write(f"```python\n")
                for msg in self.global_info["warning_msgs"]:
                    file_handler.write(msg + "\n")
                file_handler.write("```\n")
            if not self.global_info["normal_flag"]:
                file_handler.write("global-info content:\n")
                file_handler.write("```json\n")
                file_handler.write(str(self.global_info["sample_content"]) + "\n")
                file_handler.write("```\n")

        if len(self.sample_info) == 0:
            file_handler.write("数据集中无样本，请检查`sample-path`字段路径是否有效或`samples`字段是否正确" + os.linesep)
            return False

        res = self.parse_sample_info()
        total_num, normal_num, warn_num, error_num = res["total"], res["normal"], res["warn"], res["error"]
        file_handler.write("samples验证结果如下：" + os.linesep)
        res_str = f"共实例化<font color=blue>{total_num}</font>个样本，" \
                  f"其中<font color=green>{normal_num}</font>个样本实例化正常，" \
                  f"<font color=orange>{warn_num}</font>个样本出现警报信息，" \
                  f"<font color=red>{error_num}</font>个样本实例化失败。"
        file_handler.write(res_str + os.linesep)
        if total_num != normal_num:
            file_handler.write("查看具体日志信息请[点击链接](./samples_log.md)" + os.linesep)
        else:
            os.remove(osp.join(osp.split(self.output_path)[0], "samples_log.md"))
        return True

    def generate_image_info(self, file_handler):
        if not self.image_info:
            return False

        res, msgs, statis = self.parse_image_info()
        file_handler.write("## 3.部分样本可视化结果" + os.linesep)
        file_handler.write(f"共可视化<font color=blue>{len(self.image_info)}</font>个样本，"
                           f"其中<font color=green>{statis['success']}</font>个样本可视化成功，"
                           f"<font color=red>{statis['fail']}</font>个样本可视化失败" + os.linesep)
        for img in res:
            img_name = osp.split(img)[-1]
            file_handler.write(f"![{img_name}]({img})" + os.linesep)
        file_handler.write(f"具体日志信息如下：" + os.linesep + "```\n")
        for msg in msgs:
            file_handler.write(msg + "\n")
        file_handler.write("```" + os.linesep)
        return True

    def generate(self):
        with open(self.output_path, "w", encoding="utf-8") as f:
            f.write("# 数据集验证报告" + os.linesep)
            if self.parser_info:
                self.generate_parser_info(f)
            self.generate_sample_info(f)
            if self.image_info:
                self.generate_image_info(f)
        return
